/*
   ramdisk.c - A translator for ram disks

   Copyright (C) 2000 Free Software Foundation, Inc.

   Written by Neal H Walfield <neal@walfield.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111  USA
 */

#define _GNU_SOURCE 1

#include <hurd/trivfs.h>
#include <hurd/hurd_types.h>

#include <stdlib.h>
#include <argp.h>
#include <argz.h>
#include <error.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <cthreads.h>
#include <rwlock.h>

int trivfs_fstype = FSTYPE_MISC;
int trivfs_fsid = 0;

/* We provide reading and writing */
int trivfs_allow_open = O_READ | O_WRITE;
int trivfs_support_read = 1;
int trivfs_support_write = 1;
int trivfs_support_exec = 0;

/* The ram disk */
struct buffer {
  mode_t mode;               /* The permissions on the file */
  int internalsize;          /* The amount of memory allocated to the buffer */
  int size;                  /* The size of the file */
  int maxsize;               /* Maximum size that the buffer can grow to */
  int preallocated;          /* If the buffer was preallocated, also how small
                              * it can it be shrunk to */
  int initialsize;           /* The initial size of the file */
  struct rwlock lock;        /* reader / writer lock */
  void *pool;                /* The file */
};

/* By default the file is 0644 and everthing else is initialized to 0 */
static struct buffer buffer = { (S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR), };

/* The offset into the file
 * This struct is hooked onto all peropen's to keep track of the file position
 */
struct open
{
  off_t offset;
  struct mutex lock;
};

/* The options that we support
 */
static const struct argp_option options[] =
{
  {"size",         's', "INTEGER", 0, "The initial size of the file, in bytes. If this is "
                                      "larger than `maxsize' it will be reduced appropriately"},
  {"maxsize",      'm', "INTEGER", 0, "The maximum size, in bytes, that the buffer may grow "
                                      "to.  If this is zero, the buffer may grow infinitely "
                                      "large."},
  {"preallocate",  'p', "INTEGER", 0, "Allocate this amount of memory up front. If this is"
                                      "zero and the maximum size is also specified, all memory "
                                      "is preallocated"},
  {0}
};

/* Parse the options
 */
static error_t
parse_opt (int opt, char *arg, struct argp_state *state)
{
  char *endptr;

  switch (opt) {
    case 'p': /* Initial size of buffer */
      buffer.preallocated = strtol(arg, &endptr, 0);
      if (*endptr || buffer.preallocated < 0)
        argp_error(state, "Invalid argument to `preallocate'");

      if (buffer.preallocated == 0)
        buffer.preallocated = -1;
      break;

    case 'm': /* Max size of file */
      buffer.maxsize = strtol(arg, &endptr, 0);
      if (*endptr || buffer.maxsize < 0)
        argp_error(state, "Invalid argument to `maxsize'");
      break;
    case 's': /* Initial file size */
      buffer.initialsize = strtol(arg, &endptr, 0);
      if (*endptr || buffer.initialsize < 0)
        argp_error(state, "Invalid argument to `size'");
      break;
    case ARGP_KEY_INIT:
    case ARGP_KEY_SUCCESS:
    case ARGP_KEY_ERROR:
      break;

    case ARGP_KEY_END:
      /* `Fix' the arguments */
      if (buffer.maxsize && buffer.initialsize > buffer.maxsize)
        buffer.initialsize = buffer.maxsize;
      if (buffer.preallocated < buffer.initialsize)
        buffer.preallocated = buffer.initialsize;
      buffer.size = buffer.initialsize;

      /* Do some preallocation... if desired */
      switch (buffer.preallocated) {
        case 0:
          /* No need to preallocate anything */
          break;
        case -1:
           /* Pre allocate all memory */
          if (!buffer.maxsize) {
             /* Can not allocate an infinate amount */
             buffer.internalsize = 0;
             break;
          } else {
            /* Allocate it all! */
            buffer.preallocated = buffer.internalsize = buffer.maxsize;
          }
          buffer.pool = malloc(buffer.internalsize);
          break;
        default:
          if (buffer.maxsize && buffer.preallocated > buffer.maxsize)
            /* Don't want to waste memory */
            buffer.preallocated = buffer.maxsize;
          buffer.internalsize = buffer.preallocated;
          buffer.pool = malloc(buffer.internalsize);
      }
      break;

    default:
      return ARGP_ERR_UNKNOWN;
  }
  return 0;
}

static struct argp our_argp = { options, parse_opt, 0, 0 };

/* Force libtrivfs to use our argp to parse options passed in an
 * fsys_set_options RPC.
 */
struct argp *trivfs_runtime_argp = &our_argp;

/* This will be called to construct the answer to
 * fsys_get_options RPCs.
 */
kern_return_t
trivfs_append_args (struct trivfs_control *fsys,
                    char **argz, size_t *argz_len)
{
  int err = 0;
  char *options;

  if (!asprintf (&options, "--preallocate=%d --maxsize=%d --size=%d",
                 buffer.preallocated, buffer.maxsize, buffer.initialsize))
    return ENOMEM;

  err = argz_add (argz, argz_len, options);
  free (options);

  return err;
}

/* This function is heavily influenced by Gordon Matzigkeit's
 * <gord@fig.org> hello world translators (/src/hurd/trans/hello*)
 */
int main (int argc, char **argv)
{
  error_t err;
  mach_port_t bootstrap;
  struct trivfs_control *fsys;

  rwlock_init(&buffer.lock);

  /* We use the same argp for options available at startup
   * as for options we'll accept in an fsys_set_options RPC.
   */
  argp_parse (&our_argp, argc, argv, 0, 0, 0);

  task_get_bootstrap_port (mach_task_self(), &bootstrap);
  if (bootstrap == MACH_PORT_NULL)
    error (1, 0, "Must be started as a translator");

  /* Reply to our parent */
  err = trivfs_startup (bootstrap, 0, 0, 0, 0, 0, &fsys);
  mach_port_deallocate (mach_task_self(), bootstrap);
  if (err)
    error (3, err, "trivfs_startup");

  /* Launch */
  ports_manage_port_operations_multithread (fsys->pi.bucket, trivfs_demuxer,
                                            10 * 1000, /* idle thread */
                                            10 * 60 * 1000, /* idle server */
                                            0);

  return 0;
}

/*************************************************************************/

/* Return the size and the permissions of the file
 * This is called whenever the file is stat'ed, NOT when it is
 * modified.
 */
void trivfs_modify_stat(struct trivfs_protid *cred, struct stat *st)
{
  st->st_mode = ~(S_IFMT | ALLPERMS) | S_IFREG | buffer.mode;
  st->st_size = buffer.size;
}

/* Clean up and `go away'
 */
kern_return_t
trivfs_goaway(struct trivfs_control *cntl, int flags)
{
  if (!buffer.pool) free(buffer.pool);
  exit(0);
}

/* Whenever a process opens the file, we must initialize the
 * offset and save it on the hook
 */
static error_t
open_file_hook(struct trivfs_peropen *peropen)
{
  struct open *op = malloc(sizeof(struct open));
  if (!op) return ENOMEM;

  /* Append mode? */
  if (O_APPEND & peropen->openmodes)
    op->offset = buffer.size;
  else
    op->offset = 0;
  peropen->hook = op;

  mutex_init(&op->lock);

  return 0;
}

/* Whenever a file is closed, the hook can be freed
 */
static void close_file_hook(struct trivfs_peropen *peropen)
{
  struct open *op = peropen->hook;

  mutex_clear(&op->lock);
  free(peropen->hook);
}

/* Save the open and close hooks for use by libtrivfs */
error_t (*trivfs_peropen_create_hook)(struct trivfs_peropen *) = open_file_hook;
void (*trivfs_peropen_destroy_hook) (struct trivfs_peropen *) = close_file_hook;

/* How much data can be read without blocking
 */
kern_return_t
trivfs_S_io_readable (struct trivfs_protid *cred,
                      mach_port_t reply, mach_msg_type_name_t reply_type,
                      size_t *amount)
{
  error_t err;

  if (!cred)
    err = EOPNOTSUPP;
  else if (!(cred->po->openmodes & O_READ))
     err = EBADF;
  else {
    *amount = buffer.size;
    err = 0;
  }

  return err;
}

/* Read data from the buffer into data.  If offset is -1, the data
 * should be read from the current file position, otherwise, offset
 * indicates the absolute position in the file. The amount that the
 * process is interested in reading is passed in amount.  The amount
 * of data actually read into data should be stored in datalen. The
 * file position should be updated accordingly.
 */
kern_return_t
trivfs_S_io_read (struct trivfs_protid *cred,
                  mach_port_t reply, mach_msg_type_name_t reply_type,
                  vm_address_t *data, mach_msg_type_number_t *datalen,
                  off_t offset, mach_msg_type_number_t amount)
{
  struct open *op;

  /* Deny access if they have bad credentials. */
  if (!cred)
    return EOPNOTSUPP;
  else if (!(cred->po->openmodes & O_READ))
    return EBADF;

  op = cred->po->hook;

  mutex_lock(&op->lock);

  /* Get the offset. */
  if (offset == -1) {
    offset = op->offset;
  }

  rwlock_reader_lock(&buffer.lock);

  /* Prune the amount they want to read. */
  if (offset > buffer.size)
    offset = buffer.size;
  if (offset + amount > buffer.size)
    amount = buffer.size - offset;

  if (amount > 0) {
    /* Make sure that there is enough memory */
    if (*datalen < amount)
      *data = (vm_address_t) mmap(0, amount, PROT_READ|PROT_WRITE,
                   MAP_ANON, 0, 0);

    memcpy((char *)*data, buffer.pool + offset, amount);

    rwlock_reader_unlock(&buffer.lock);

    /* Change the file position */
    op->offset += amount;
  } else {
    rwlock_reader_unlock(&buffer.lock);
  }

  mutex_unlock(&op->lock);

  *datalen = amount;
  return 0;
}

/* Block until one of the the possible operations, SELECT_READ,
 * SELECT_WRITE, or SELECT_URG or'ed in select can be completed
 * without blocking. Store the result in select.
 */
error_t
trivfs_S_io_select (struct trivfs_protid *cred,
                    mach_port_t reply, mach_msg_type_name_t reply_type,
                    int *select, int *tag)
{
  if (!cred)
    return EOPNOTSUPP;
  if (((*select & SELECT_READ) && !(cred->po->openmodes & O_READ)) ||
      ((*select & SELECT_WRITE) && !(cred->po->openmodes & O_WRITE)))
    return EBADF;

  return 0;
}

/* Seek to a point in the file.  If whence is SEEK_SET offset indicates
 * a position relative to the start of the file.  If whence is SEEK_CUR,
 * offset is relative to the current file position.  If whence is
 * SEEK_END, offset is relative to the end of the file.
 */
kern_return_t
trivfs_S_io_seek(struct trivfs_protid *cred,
                  mach_port_t reply, mach_msg_type_name_t reply_type,
                  off_t offset, int whence, off_t *new_offset)
{
  struct open *op;

  if (!cred)
    return EOPNOTSUPP;

  op = cred->po->hook;

  mutex_lock(&op->lock);

  switch (whence) {
    case SEEK_SET:
      op->offset = offset;
      break;
    case SEEK_CUR:
      op->offset += offset;
      break;
    case SEEK_END:
      op->offset = buffer.size - offset;
      break;
    default:
      return EINVAL;
  }

  /* Return the new offset */
  *new_offset = op->offset;

  mutex_unlock(&op->lock);

  return 0;
}

/* Sets the size of the buffer to a given size.  When entering,
 * one must already have a writer lock on buffer.lock and have
 * op->lock; also, op MUST be valid! This is called by
 * trivfs_S_file_set_size and trivfs_S_io_write.
 */
static kern_return_t
set_size(struct open *op, off_t size)
{
  int i;

  /* Don't go over the maximim size */
  if (buffer.maxsize && (size > buffer.maxsize))
    size = buffer.maxsize;

  /* Is the data being written past the EOF? */
  if (size > buffer.size) {
    /* Do we need to allocate more memory */
    if (size > buffer.internalsize) {
      /* Allocate size and a bit extra but only if it is less
       * then the max size
       */
      i = (size + (size >> 3));
      if (i < 1024) i = 1024; /* Try at least 1k */
      if (buffer.maxsize && i > buffer.maxsize) i = buffer.maxsize;

      /* Try the allocation */
      buffer.pool = realloc(buffer.pool, i);

      /* Was the allocation successful? */
      if (buffer.pool) {
        buffer.internalsize = i;
      } else {
        /* Try again, without the extra memory */
        buffer.pool = realloc(buffer.pool, size);
        if (!buffer.pool)
          /* Ouch! That's it! */
          return ENOMEM;
        buffer.internalsize = size;
      }
    }
  }

  buffer.size = size;

  /* Do we have extra memory that we can give back?
   * If the internal size > 1.125 * actual size then
   * shrink it down to that size BUT only if it is not
   * below the the low water mark.
   */
  if (buffer.internalsize > (i = (buffer.size >> 3) + buffer.size) && (i > 1024)) {
    /* Don't go below the low water mark */
    if (i < buffer.preallocated)
      i = buffer.preallocated;
    if (i < buffer.internalsize) {
      buffer.internalsize = i;
      buffer.pool = realloc(buffer.pool, buffer.internalsize);
    }
  }

  /* Save the position */
  op->offset = size;

  return 0;
}

/* Set the file size to size
 */
kern_return_t
trivfs_S_file_set_size (struct trivfs_protid *cred, mach_port_t reply,
			mach_msg_type_name_t reply_type,
			off_t size)
{
  int err;
  struct open *op;

  if (!cred)
    return EOPNOTSUPP;

  op = cred->po->hook;

  mutex_lock(&op->lock);
  rwlock_writer_lock(&buffer.lock);

  err = set_size(op, size);

  rwlock_writer_unlock(&buffer.lock);
  mutex_unlock(&op->lock);

  return err;
}

/* Write datalen bytes of data stored in data into the buffer.  If
 * offset is -1, the data is written according to the current file
 * position, otherwise, the offset indicates the absolute position.
 * The amount of data actually written should be stored in amount.
 * The current file position should be updated accordingly.
 */
kern_return_t
trivfs_S_io_write (struct trivfs_protid *cred, mach_port_t reply,
                   mach_msg_type_name_t reply_type, char *data, int datalen,
                   off_t offset, int *amount)
{
  struct open *op;
  int err = 0;

  if (!cred)
    return EOPNOTSUPP;
  else if (!(cred->po->openmodes & O_WRITE))
    return EBADF;

  op = cred->po->hook;

  mutex_lock(&op->lock);

  /* If offset is -1 then write at the current position */
  if (offset == -1) {
    offset = op->offset;
  }

  rwlock_writer_lock(&buffer.lock);

  /* Do we need to resize the buffer? */
  if (offset + datalen > buffer.size) {
    /* Would we go over the maximum allowed size? */
    if (buffer.maxsize && (offset + datalen > buffer.maxsize)) {
      err = set_size(op, buffer.maxsize);
      if (!err) {
        /* Did we fail to allocate it all? */
        if (offset + datalen > buffer.size)
          datalen = buffer.size - offset;
        /* Are we now at EOF? */
        if (datalen <= 0)
          err = ENOSPC;
      }
    } else {
      /* We have an inifinite amount of space to grow, if needed */
      err = set_size(op, offset + datalen);
    }
  }

  if (err) {
    rwlock_writer_unlock(&buffer.lock);
    mutex_unlock(&op->lock);
    return err;
  }

  /* Copy it in */
  memcpy(buffer.pool + offset, data, datalen);

  rwlock_writer_unlock(&buffer.lock);

  /* Update the offset */
  op->offset = offset + datalen;
  mutex_unlock(&op->lock);

  /* Note how much was copied */
  *amount = datalen;

  return 0;
}

/* Modify the permissions of the file.  If we get this far, then the kernel
 * must think that it is ok to do so
 */
kern_return_t
trivfs_S_file_chmod (struct trivfs_protid *cred,
                     mach_port_t reply, mach_msg_type_name_t reply_type,
                     mode_t mode)
{
  if (!cred)
    return EOPNOTSUPP;

  /* There is no need to lock this since we are the only function which
   * writes to this and nobody actually `operates' on the value.
   */
  buffer.mode = mode;
  return 0;
}

/* The following four routines modify the open mode, we don't
 * actually use these post open, so we should be able to safely
 * ignore them.
 */
kern_return_t
trivfs_S_io_get_openmodes (struct trivfs_protid *cred,
                           mach_port_t reply, mach_msg_type_name_t replytype,
                           int *bits)
{
  if (!cred)
    return EOPNOTSUPP;
  else {
    *bits = cred->po->openmodes;
    return 0;
  }
}

kern_return_t
trivfs_S_io_set_all_openmodes(struct trivfs_protid *cred,
                              mach_port_t reply,
                              mach_msg_type_name_t replytype,
                              int mode)
{
  if (!cred)
    return EOPNOTSUPP;
  else
    return 0;
}

kern_return_t
trivfs_S_io_set_some_openmodes (struct trivfs_protid *cred,
                                mach_port_t reply,
                                mach_msg_type_name_t replytype,
                                int bits)
{
  if (!cred)
    return EOPNOTSUPP;
  else
    return 0;
}

kern_return_t
trivfs_S_io_clear_some_openmodes (struct trivfs_protid *cred,
                                  mach_port_t reply,
                                  mach_msg_type_name_t replytype,
                                  int bits)
{
  if (!cred)
    return EOPNOTSUPP;
  else
    return 0;
}

